#!/usr/bin/env python3

import argparse
import asyncio
import logging
import mimetypes
import os.path

import aiotus


def aiotus_upload() -> int:
    parser = argparse.ArgumentParser(
        description="Tool to upload a file to a tus (tus.io) server.",
        epilog="This program is part of the aiotus python package.",
    )
    parser.add_argument("--debug", action="store_true", help="log debug messages")
    parser.add_argument("endpoint", type=str, help="creation URL of the tus server")
    parser.add_argument("file", type=str, help="file to upload")
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    metadata = {"filename": os.path.basename(args.file).encode()}

    mime_type, _ = mimetypes.guess_type(args.file)
    if mime_type:
        metadata["mime_type"] = mime_type.encode()

    try:
        with open(args.file, "rb") as file:
            location = asyncio.run(aiotus.upload(args.endpoint, file, metadata))
            if location:
                print(str(location))
                return 0
    except KeyboardInterrupt:  # pragma: no cover
        pass
    except Exception as e:
        logging.error(f"Unable to upload file: {e}")

    return 1


def aiotus_metadata() -> int:
    parser = argparse.ArgumentParser(
        description="Tool to query the metadata of a file on a tus (tus.io) server.",
        epilog="This program is part of the aiotus python package.",
    )
    parser.add_argument("--debug", action="store_true", help="log debug messages")
    parser.add_argument("location", type=str, help="file location on the tus server")
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    try:
        metadata = asyncio.run(aiotus.metadata(args.location))
        assert isinstance(metadata, dict)  # nosec - Silence mypy.

        for k, v in metadata.items():
            if v is None:
                print(f"{k}")
            else:
                value = repr(v)[2:][:-1]
                print(f"{k}: {value}")

        return 0
    except KeyboardInterrupt:  # pragma: no cover
        return 1



import dataclasses
import io
import os.path
import tempfile
from typing import Optional

import aiohttp
import pytest  # type: ignore
import yarl
from xprocess import ProcessStarter  # type: ignore


@pytest.fixture
async def tus_server(aiohttp_server):
    """Return a fake tus server that can consume a single file."""

    state = {
        # Number of times the respective handlers will fail.
        "retries_create": 0,
        "retries_head": 0,
        "retries_upload": 0,
        # URLs to create and upload. (Filled out later.)
        "create_endpoint": None,
        "upload_endpoint": None,
        # The uploaded data will be accumulated here.
        "data": None,
        # Metadata included in the creation will be placed here.
        "metadata": None,
        # Complete HTTP headers used in the last head/post request.
        "head_headers": None,
        "post_headers": None,
    }

    async def handler_create(request):
        state["retries_create"] -= 1
        if state["retries_create"] > 0:
            raise aiohttp.web.HTTPInternalServerError()

        if "Upload-Metadata" in request.headers:
            state["metadata"] = request.headers["Upload-Metadata"]

        state["post_headers"] = request.headers

        # "Create" the upload.
        state["data"] = bytearray()

        headers = {"Location": str(state["upload_endpoint"])}
        raise aiohttp.web.HTTPCreated(headers=headers)

    async def handler_head(request):
        state["retries_head"] -= 1
        if state["retries_head"] > 0:
            raise aiohttp.web.HTTPInternalServerError()

        state["head_headers"] = request.headers

        if state["data"] is None:
            raise aiohttp.web.HTTPNotFound()

        headers = {"Upload-Offset": str(len(state["data"]))}
        if state["metadata"] is not None:
            headers["Upload-Metadata"] = state["metadata"]
        raise aiohttp.web.HTTPOk(headers=headers)

    async def handler_upload(request):
        body = await request.read()

        state["retries_upload"] -= 1
        if state["retries_upload"] > 0:
            # Pretend we did only receive half of the data before an error happend.
            state["data"].extend(body[: len(body) // 2])
            raise aiohttp.web.HTTPInternalServerError()

        state["data"].extend(body)
        raise aiohttp.web.HTTPNoContent()

    upload_name = "1234abcdefgh"

    app = aiohttp.web.Application()
    app.router.add_route("POST", "/files", handler_create)
    app.router.add_route("HEAD", "/files/" + upload_name, handler_head)
    app.router.add_route("PATCH", "/files/" + upload_name, handler_upload)

    state["server"] = await aiohttp_server(app)
    state["create_endpoint"] = state["server"].make_url("/files")
    state["upload_endpoint"] = state["create_endpoint"] / upload_name

    return state


@pytest.fixture
def memory_file():
    """Dummy data to use during tests."""
    return io.BytesIO(b"\x00\x01\x02\x03")


@dataclasses.dataclass
class TusServer:

    # The URL where the server is listening on.
    url: yarl.URL

    # The path of the certificate file of the server, if TLS is used.
    certificate: Optional[str] = None


@pytest.fixture(scope="module")
def tusd(pytestconfig, xprocess):
    """Start the tusd (tus.io reference implementation) and yield the upload URL.

    Assumes that the tusd executable is located in the pytest rootdir.
    """

    host = "0.0.0.0"
    port = "8080"
    basepath = "/files/"

    executable = pytestconfig.rootdir.join("/tusd")

    class Starter(ProcessStarter):
        pattern = "You can now upload files to:"
        args = [executable, "-host", host, "-port", port, "-base-path", basepath]

    server_name = "tusd-server"

    xprocess.ensure(server_name, Starter)
    server = TusServer(yarl.URL(f"http://{host}:{port}{basepath}"))
    yield server
    xprocess.getinfo(server_name).terminate()


# Template for the nginx configuration file. Stripped-down from
# https://github.com/tus/tusd/blob/master/examples/nginx.conf
_nginx_conf = """
    pid nginx.pid;
    daemon off;
    error_log /dev/null;

    events {{
    }}

    http {{
        client_body_temp_path ./client_body;
        proxy_temp_path ./proxy;
        fastcgi_temp_path ./fastcgi;
        uwsgi_temp_path ./uwsgi;
        scgi_temp_path ./scgi;

        access_log off;

        server {{
            listen {port} ssl;
            server_name _;
            ssl_certificate {crt};
            ssl_certificate_key {key};

            location / {{
                # Forward incoming requests to local tusd instance
                proxy_pass {tusd_url};

                # Disable request and response buffering
                proxy_request_buffering  off;
                proxy_buffering off;
                proxy_http_version 1.1;

                # Add X-Forwarded-* headers
                proxy_set_header X-Forwarded-Host $hostname;
                proxy_set_header X-Forwarded-Proto $scheme;

                proxy_set_header Upgrade $http_upgrade;
                proxy_set_header Connection "upgrade";
                client_max_body_size 0;
            }}
        }}
    }}
"""


@pytest.fixture(scope="module")
def nginx_proxy(xprocess, tusd):
    """Start an nginx proxy in front of tusd that does TLS termination."""

    test_dir = os.path.dirname(os.path.abspath(__file__))
    fmt = {
        "crt": os.path.join(test_dir, "selfsigned.crt"),
        "key": os.path.join(test_dir, "nginx.key"),
        "port": 8443,
        "tusd_url": str(tusd.url),
    }
    conf = _nginx_conf.format(**fmt)

    with tempfile.TemporaryDirectory() as d:
        conf_file = os.path.join(d, "nginx.conf")
        with open(conf_file, "w") as f:
            f.write(conf)

        class Starter(ProcessStarter):
            pattern = "could not open error log file"
            args = ["nginx", "-p", d, "-c", "nginx.conf"]

        server_name = "nginx-server"

        xprocess.ensure(server_name, Starter)
        server = TusServer(yarl.URL(f"https://localhost:{fmt['port']}"))
        server.certificate = fmt["crt"]
        yield server
        xprocess.getinfo(server_name).terminate()



"""
Implementation of the
`core tus protocol <https://tus.io/protocols/resumable-upload.html#core-protocol>`_.
"""

import asyncio
import base64
import io
from copy import copy
from typing import BinaryIO, Dict, Optional

import aiohttp
import yarl

from . import common
from .log import logger


async def offset(
    session: aiohttp.ClientSession,
    location: yarl.URL,
    ssl: common.SSLArgument = None,
    headers: Optional[Dict[str, str]] = None,
) -> int:
    """Get the number of uploaded bytes.

    :param session: HTTP session to use for connections.
    :param location: The upload endpoint to query.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The number of bytes that are already on the server.
    """

    headers = headers or {}
    headers["Tus-Resumable"] = common.TUS_PROTOCOL_VERSION

    logger.debug(f'Getting offset of "{location}"...')
    async with await session.head(location, headers=headers, ssl=ssl) as response:
        response.raise_for_status()

        if "Upload-Offset" not in response.headers:
            raise RuntimeError(
                f"HEAD request succeeded for {location}, "
                'but no "Upload-Offset" header in response.'
            )

        try:
            offset_header = response.headers["Upload-Offset"]
            current_offset = int(offset_header)
            if current_offset < 0:
                raise RuntimeError()

            return current_offset
        except asyncio.CancelledError:  # pragma: no cover
            raise
        except Exception:
            raise RuntimeError(
                f'Unable to convert "Upload-Offset" header '
                f'"{offset_header}" to a positive integer.'
            )


def _parse_metadata(header: str) -> common.Metadata:
    """Split and decode the input into a metadata dictionary."""

    header = header.strip()
    if not header:
        return {}

    md: common.Metadata = {}
    for pair in header.split(","):
        kv = pair.split()
        if len(kv) == 1:
            md[kv[0]] = None
        elif len(kv) == 2:
            md[kv[0]] = base64.b64decode(kv[1], validate=True)
        else:
            raise ValueError("Key/Value pair consists of more than two elements.")

    return md


async def metadata(
    session: aiohttp.ClientSession,
    location: yarl.URL,
    ssl: common.SSLArgument = None,
    headers: Optional[Dict[str, str]] = None,
) -> common.Metadata:
    """Get the metadata associated with an upload.

    :param session: HTTP session to use for connections.
    :param location: The upload endpoint to query.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The metadata of the upload.
    """

    headers = headers or {}
    headers["Tus-Resumable"] = common.TUS_PROTOCOL_VERSION

    logger.debug(f'Getting metadata of "{location}"...')
    async with await session.head(location, headers=headers, ssl=ssl) as response:
        response.raise_for_status()

        if "Upload-Metadata" not in response.headers:
            return {}

        try:
            return _parse_metadata(response.headers["Upload-Metadata"])
        except asyncio.CancelledError:  # pragma: no cover
            raise
        except Exception as e:
            raise common.ProtocolError(f"Unable to parse metadata: {e}")


async def upload_buffer(
    session: aiohttp.ClientSession,
    location: yarl.URL,
    buffer: BinaryIO,
    ssl: common.SSLArgument = None,
    chunksize: int = 4 * 1024 * 1024,
    headers: Optional[Dict[str, str]] = None,
) -> None:
    """Upload data to the server.

    :param session: HTTP session to use for connections.
    :param location: The endpoint to upload to.
    :param buffer: The data to upload.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param chunksize: The size of individual chunks to upload at a time.
    :param headers: Optional headers used in the request.
    """

    loop = asyncio.get_event_loop()
    headers = headers or {}

    current_offset = await offset(session, location, ssl=ssl)
    await loop.run_in_executor(None, buffer.seek, current_offset, io.SEEK_SET)

    logger.debug(f'Resuming upload of "{location}" at offset {current_offset}..."')

    while True:
        chunk = await loop.run_in_executor(None, buffer.read, chunksize)
        if not chunk:
            logger.debug("EOF reached")
            break

        # Use the parameter headers as base to make sure that users do not overwrite
        # important data like 'Content-Type'.
        request_headers = copy(headers)
        request_headers.update(
            {
                "Tus-Resumable": common.TUS_PROTOCOL_VERSION,
                "Upload-Offset": str(current_offset),
                "Content-Length": str(len(chunk)),
                "Content-Type": "application/offset+octet-stream",
            }
        )

        logger.debug(f'Uploading {len(chunk)} bytes to "{location}"...')
        async with await session.patch(
            location, headers=request_headers, data=chunk, ssl=ssl
        ) as response:
            response.raise_for_status()

        current_offset += len(chunk)



import ssl
from typing import Dict, Optional, Union

import aiohttp

# The version of the tus protocol we implement.
TUS_PROTOCOL_VERSION = "1.0.0"

#: The type of the 'ssl' argument passed on to aiohttp.
SSLArgument = Optional[Union[bool, ssl.SSLContext, aiohttp.Fingerprint]]

#: The type used to hold metadata of an upload.
Metadata = Dict[str, Optional[bytes]]


class ProtocolError(Exception):
    """Server response did not follow the tus protocol."""



"""
Implementation of the
`creation extension <https://tus.io/protocols/resumable-upload.html#creation>`_
to the tus protocol.
"""

import asyncio
import base64
import io
from typing import BinaryIO, Dict, Optional

import aiohttp
import yarl

from . import common
from .log import logger


async def create(
    session: aiohttp.ClientSession,
    url: yarl.URL,
    file: BinaryIO,
    metadata: common.Metadata,
    ssl: common.SSLArgument = None,
    headers: Optional[Dict[str, str]] = None,
) -> yarl.URL:
    """Create an upload.

    :param session: HTTP session to use for connections.
    :param url: The creation endpoint of the server.
    :param file: The file object to upload.
    :param metadata: Additional metadata for the upload.
    :param ssl: SSL validation mode, passed on to aiohttp.
    :param headers: Optional headers used in the request.
    :return: The URL to upload the data to.
    """

    loop = asyncio.get_event_loop()
    total_size = await loop.run_in_executor(None, file.seek, 0, io.SEEK_END)
    headers = headers or {}

    headers.update(
        {
            "Tus-Resumable": common.TUS_PROTOCOL_VERSION,
            "Upload-Length": str(total_size),
        }
    )

    if metadata:
        # Check metadata keys before we proceed.
        for k in metadata:
            if not k.isascii():
                raise ValueError("Metadata keys must only contain ASCII characters.")

            if " " in k:
                raise ValueError("Metadata keys must not contain spaces.")

            if "," in k:
                raise ValueError("Metadata keys must not contain commas.")

        def encode_value(value: Optional[bytes]) -> str:
            if value is None:
                return ""

            encoded_bytes = base64.b64encode(value)
            encoded_string = encoded_bytes.decode()
            return " " + encoded_string

        pairs = [f"{k}{encode_value(v)}" for k, v in metadata.items()]
        headers["Upload-Metadata"] = ", ".join(pairs)

    logger.debug("Creating upload...")
    async with await session.post(url, headers=headers, ssl=ssl) as response:
        if response.status != 201:
            raise common.ProtocolError(
                f"Wrong status code {response.status}, expected 201."
            )

        if "Location" not in response.headers:
            raise RuntimeError('Upload created, but no "Location" header in response.')

        return yarl.URL(response.headers["Location"])



"""Test the implementation of the core protocol."""

import binascii

import aiohttp
import pytest  # type: ignore

import aiotus


class TestOffset:
    async def test_offset_exceptions(self, aiohttp_server, memory_file):
        """Check for the different exceptions that can be thrown."""

        async def handler_not_found(request):
            raise aiohttp.web.HTTPNotFound()

        async def handler_no_offset(request):
            raise aiohttp.web.HTTPOk()

        async def handler_wrong_offset(request):
            headers = {"Upload-Offset": "xyz"}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_negative_offset(request):
            headers = {"Upload-Offset": "-1"}
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/not_found", handler_not_found)
        app.router.add_route("HEAD", "/no_offset", handler_no_offset)
        app.router.add_route("HEAD", "/wrong_offset", handler_wrong_offset)
        app.router.add_route("HEAD", "/negative_offset", handler_negative_offset)
        server = await aiohttp_server(app)

        # Check if status code is checked correctly.
        with pytest.raises(aiohttp.ClientResponseError):
            location = server.make_url("/not_found")

            async with aiohttp.ClientSession() as session:
                await aiotus.core.offset(session, location)

        # Check if the check for the "Upload-Offset" header is working.
        with pytest.raises(RuntimeError) as excinfo:
            location = server.make_url("/no_offset")

            async with aiohttp.ClientSession() as session:
                await aiotus.core.offset(session, location)

        assert 'no "Upload-Offset" header' in str(excinfo.value)

        # Check if the offset value is checked properly.
        with pytest.raises(RuntimeError) as excinfo:
            location = server.make_url("/wrong_offset")

            async with aiohttp.ClientSession() as session:
                await aiotus.core.offset(session, location)

        assert 'Unable to convert "Upload-Offset" header' in str(excinfo.value)

        with pytest.raises(RuntimeError) as excinfo:
            location = server.make_url("/negative_offset")

            async with aiohttp.ClientSession() as session:
                await aiotus.core.offset(session, location)

        assert 'Unable to convert "Upload-Offset" header' in str(excinfo.value)

    async def test_offset_functional(self, aiohttp_server):
        """Test the normal functionality of the '_offset' function."""

        async def handler(request):

            assert "Tus-Resumable" in request.headers
            assert (
                request.headers["Tus-Resumable"] == aiotus.common.TUS_PROTOCOL_VERSION
            )

            headers = {"Upload-Offset": "123"}
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/files/12345678", handler)
        server = await aiohttp_server(app)

        location = server.make_url("/files/12345678")

        async with aiohttp.ClientSession() as session:
            offset = await aiotus.core.offset(session, location)

        assert offset == 123


class TestMetadata:
    def test_parse_metadata(self):
        """Check if metadata is parsed correctly."""

        md = aiotus.core._parse_metadata("")
        assert md == {}

        md = aiotus.core._parse_metadata("key")
        assert md == {"key": None}

        md = aiotus.core._parse_metadata("key ")
        assert md == {"key": None}

        md = aiotus.core._parse_metadata("key dmFsdWU=")
        assert md == {"key": b"value"}

        md = aiotus.core._parse_metadata("k1, k2 dmFsdWU=")
        assert md == {"k1": None, "k2": b"value"}

        md = aiotus.core._parse_metadata("k1 djE=, k2 djI=  ")
        assert md == {"k1": b"v1", "k2": b"v2"}

        with pytest.raises(binascii.Error) as excinfo:
            md = aiotus.core._parse_metadata("k1 djE")
        assert "padding" in str(excinfo.value)

        with pytest.raises(binascii.Error) as excinfo:
            md = aiotus.core._parse_metadata("k1 dj&=")
        assert "Non-base64" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            md = aiotus.core._parse_metadata("k v v")
        assert "more than two elements" in str(excinfo.value)

    async def test_metadata(self, aiohttp_server):
        """Check the 'core.metadata()' function."""

        async def handler_no_metadata(request):
            headers = {"Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION}
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_invalid(request):
            headers = {
                "Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION,
                "Upload-Metadata": "k1 djE",
            }
            raise aiohttp.web.HTTPOk(headers=headers)

        async def handler_valid(request):
            headers = {
                "Tus-Resumable": aiotus.common.TUS_PROTOCOL_VERSION,
                "Upload-Metadata": "k1, k2 dmFsdWU=",
            }
            raise aiohttp.web.HTTPOk(headers=headers)

        app = aiohttp.web.Application()
        app.router.add_route("HEAD", "/no_metadata", handler_no_metadata)
        app.router.add_route("HEAD", "/invalid", handler_invalid)
        app.router.add_route("HEAD", "/valid", handler_valid)
        server = await aiohttp_server(app)

        async with aiohttp.ClientSession() as session:
            md = await aiotus.core.metadata(session, server.make_url("/no_metadata"))
            assert md == {}

            with pytest.raises(aiotus.ProtocolError) as excinfo:
                md = await aiotus.core.metadata(session, server.make_url("/invalid"))
                assert "Unable to parse metadata" in str(excinfo.value)

            md = await aiotus.core.metadata(session, server.make_url("/valid"))
            assert md == {"k1": None, "k2": b"value"}



import os
import re
import sys


aiotus_path = os.path.join(os.path.dirname(__file__), '../..')
aiotus_path = os.path.abspath(aiotus_path)
sys.path.insert(0, aiotus_path)


def get_version():
    """Return package version from setup.py (hacky)."""

    try:
        filename = os.path.join(os.path.dirname(__file__), '../..', 'setup.py')
        with open(filename, 'r') as fd:
            setup_py = fd.read()

        m = re.search(r'version="(\d+\.\d+\.\d+)"', setup_py)
        return m.group(1)
    except:
        sys.exit('Unable to get package version from setup.py.')


project = 'aiotus'
copyright = '2020, Jens Steinhauser'
author = 'Jens Steinhauser'
release = get_version()

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'aiohttp': ('https://aiohttp.readthedocs.io/en/stable/', None),
    'yarl': ('https://yarl.readthedocs.io/en/stable/', None),
}

exclude_patterns = []
html_static_path = ['_static']
html_theme = 'alabaster'
html_theme_options = {
    'description': 'Asynchronous client side implementation of the tus protocol for Python.',
}
templates_path = ['_templates']



import asyncio
import contextlib
import dataclasses
from typing import (
    AsyncContextManager,
    AsyncIterator,
    BinaryIO,
    Callable,
    Dict,
    Optional,
    Union,
)

import aiohttp
import tenacity  # type: ignore
import yarl

from . import common, core, creation
from .log import logger


@contextlib.asynccontextmanager
async def asyncnullcontext(
    enter_result: aiohttp.ClientSession,
) -> AsyncIterator[aiohttp.ClientSession]:
    """Asynchronous version of 'contextlib.nullcontext()'.

    (Typed to 'aiohttp.ClientSession', as I could not get 'AsyncIterator[]' work with
    generic 'TypeVar's.)
    """
    yield enter_result


@dataclasses.dataclass
class RetryConfiguration:
    """Class to hold settings for the functions of this module."""

    retry_attempts: int = 10
    """
    Number of retry attempts to do when the communication fails.
    """

    max_retry_period_seconds: float = 60.0
    """
    Maximum time between retries, in seconds.

    Exponential backoff is used in case of communication errors,
    but the time between retries is caped by this value.
    """

    ssl: common.SSLArgument = None
    """
    'ssl' argument passed on to the aiohttp calls.

    This can be None, False, or an instance of ssl.SSLContext, see
    the `aiohttp documentation
    <https://docs.aiohttp.org/en/stable/client_advanced.html#ssl-control-for-tcp-sockets>`_
    for the different meanings.
    """  # noqa: E501


def _make_log_before_function(s: str) -> Callable[[tenacity.RetryCallState], None]:
    def log(retry_state: tenacity.RetryCallState) -> None:

        if retry_state.attempt_number > 1:
            logger.info(
                f"Trying {s} again, attempt number {retry_state.attempt_number}..."
            )

    return log


def _make_log_before_sleep_function(
    s: str,
) -> Callable[[tenacity.RetryCallState], None]:
    def log(retry_state: tenacity.RetryCallState) -> None:
        if (retry_state.next_action is not None) and (retry_state.outcome is not None):
            duration = retry_state.next_action.sleep
            if retry_state.outcome.failed:
                value = retry_state.outcome.exception()
            else:
                value = retry_state.outcome.result()
            logger.warning(
                f"{s.capitalize()} failed, "
                f"retrying in {duration:.0f} second(s): {value}"
            )

    return log


async def upload(
    endpoint: Union[str, yarl.URL],
    file: BinaryIO,
    metadata: Optional[common.Metadata] = None,
    client_session: Optional[aiohttp.ClientSession] = None,
    config: RetryConfiguration = RetryConfiguration(),
    headers: Optional[Dict[str, str]] = None,
) -> Optional[yarl.URL]:
    """Upload a file to a tus server.

    In case of a communication error, this function retries the upload.

    :param endpoint: The creation endpoint of the server.
    :param file: The file to upload.
    :param metadata: Additional metadata for the upload.
    :param client_session: An aiohttp ClientSession to use.
    :param config: Settings to customize the retry behaviour.
    :param headers: Optional headers used in the request.
    :return: The location where the file was uploaded to (if the upload succeeded).
    """

    url = yarl.URL(endpoint)

    if metadata is None:
        metadata = {}

    retrying_create = tenacity.AsyncRetrying(
        retry=tenacity.retry_if_exception_type(aiohttp.ClientError),
        stop=tenacity.stop_after_attempt(config.retry_attempts),
        wait=tenacity.wait_exponential(max=config.max_retry_period_seconds),
        before=_make_log_before_function("upload creation"),
        before_sleep=_make_log_before_sleep_function("upload creation"),
    )

    retrying_upload_file = tenacity.AsyncRetrying(
        retry=tenacity.retry_if_exception_type(aiohttp.ClientError),
        stop=tenacity.stop_after_attempt(config.retry_attempts),
        wait=tenacity.wait_exponential(max=config.max_retry_period_seconds),
        before=_make_log_before_function("upload"),
        before_sleep=_make_log_before_sleep_function("upload"),
    )

    try:
        ctx: Union[aiohttp.ClientSession, AsyncContextManager[aiohttp.ClientSession]]
        if client_session is None:
            ctx = aiohttp.ClientSession()
        else:
            ctx = asyncnullcontext(client_session)

        async with ctx as session:
            location: yarl.URL
            location = await retrying_create.call(
                creation.create,
                session,
                url,
                file,
                metadata,
                ssl=config.ssl,
                headers=headers,
            )
            if not location.is_absolute():
                location = url / location.path

            await retrying_upload_file.call(
                core.upload_buffer,
                session,
                location,
                file,
                ssl=config.ssl,
                headers=headers,
            )

            return location
    except asyncio.CancelledError:  # pragma: no cover
        raise
    except tenacity.RetryError as e:
        logger.error(
            f"Unable to upload file, even after retrying: {e.last_attempt.exception()}"
        )
    except Exception as e:
        logger.error(f"Unable to upload file: {e}")

    return None


async def metadata(
    endpoint: Union[str, yarl.URL],
    client_session: Optional[aiohttp.ClientSession] = None,
    config: RetryConfiguration = RetryConfiguration(),
    headers: Optional[Dict[str, str]] = None,
) -> Optional[common.Metadata]:
    """Retrieve the metadata of an upload.

    In case of a communication error, this function retries.

    :param endpoint: The location of the upload.
    :param client_session: An aiohttp ClientSession to use.
    :param config: Settings to customize the retry behaviour.
    :param headers: Optional headers used in the request.
    :return: The metadata associated with the upload.
    """

    if isinstance(endpoint, str):
        url = yarl.URL(endpoint)
    else:
        url = endpoint

    retrying_metadata = tenacity.AsyncRetrying(
        retry=tenacity.retry_if_exception_type(aiohttp.ClientError),
        stop=tenacity.stop_after_attempt(config.retry_attempts),
        wait=tenacity.wait_exponential(max=config.max_retry_period_seconds),
        before=_make_log_before_function("query metadata"),
        before_sleep=_make_log_before_sleep_function("query metadata"),
    )

    try:
        ctx: Union[aiohttp.ClientSession, AsyncContextManager[aiohttp.ClientSession]]
        if client_session is None:
            ctx = aiohttp.ClientSession()
        else:
            ctx = asyncnullcontext(client_session)

        async with ctx as session:
            md: common.Metadata
            md = await retrying_metadata.call(
                core.metadata, session, url, ssl=config.ssl, headers=headers
            )

            return md
    except asyncio.CancelledError:  # pragma: no cover
        raise
    except tenacity.RetryError as e:
        logger.error(
            f"Unable to get metadata, even after retrying: {e.last_attempt.exception()}"
        )

    return None



"""Test the 'upload()' function."""

import aiohttp
import yarl

import aiotus


class TestRetry:
    async def test_upload_functional(self, tus_server, memory_file):
        """Test the normal functionality of the 'upload()' function."""

        metadata = {"Content-Type": "image/jpeg".encode(), "key": None}
        additional_headers = {"h1": "v1", "h2": "v2"}

        location = await aiotus.upload(
            tus_server["create_endpoint"],
            memory_file,
            metadata,
            headers=additional_headers,
        )

        assert location is not None
        assert tus_server["metadata"] == "Content-Type aW1hZ2UvanBlZw==, key"
        assert tus_server["data"] is not None
        assert tus_server["data"] == memory_file.getbuffer()
        assert tus_server["post_headers"] is not None
        assert "h1" in tus_server["post_headers"]
        assert tus_server["post_headers"]["h1"] == "v1"
        assert "h2" in tus_server["post_headers"]
        assert tus_server["post_headers"]["h2"] == "v2"

    async def test_upload_client_session(self, tus_server, memory_file):
        """Use a custom client session."""

        headers = {"Authorization": "Basic xyz"}
        md1 = {"key1": "value1".encode(), "key2": "value2".encode()}

        async with aiohttp.ClientSession(headers=headers) as s:
            location = await aiotus.upload(
                tus_server["create_endpoint"], memory_file, md1, client_session=s
            )

            additional_headers = {"h1": "v1", "h2": "v2"}
            md2 = await aiotus.metadata(
                location, client_session=s, headers=additional_headers
            )

            assert not s.closed

        assert tus_server["data"] is not None
        assert tus_server["data"] == memory_file.getbuffer()
        assert tus_server["post_headers"] is not None
        assert "Authorization" in tus_server["post_headers"]
        assert tus_server["post_headers"]["Authorization"] == headers["Authorization"]
        assert tus_server["head_headers"] is not None
        assert "h1" in tus_server["head_headers"]
        assert tus_server["head_headers"]["h1"] == "v1"
        assert "h2" in tus_server["head_headers"]
        assert tus_server["head_headers"]["h2"] == "v2"
        assert md1 == md2

    async def test_upload_wrong_metadata(self, tus_server, memory_file):
        """Test if wrong metadata is rejected."""

        metadata = {"²": "2".encode()}

        location = await aiotus.upload(
            tus_server["create_endpoint"], memory_file, metadata
        )

        assert location is None

    async def test_retry(self, tus_server, memory_file):
        """Test the retry functionality."""

        # Make the server fail a few times to test the retry logic.
        tus_server["retries_head"] = 3
        tus_server["retries_upload"] = 3

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)
        md1 = {"key1": "value1".encode(), "key2": "value2".encode()}

        location = await aiotus.upload(
            tus_server["create_endpoint"], memory_file, md1, config=config
        )

        assert location is not None
        assert tus_server["data"] is not None
        assert tus_server["data"] == memory_file.getbuffer()

        tus_server["retries_head"] = 3
        md2 = await aiotus.metadata(location, config=config)

        assert md1 == md2

        tus_server["retries_head"] = 11
        md = await aiotus.metadata(location, config=config)

        assert md is None

    async def test_upload_create_retries_exceeded(self, tus_server, memory_file):
        """Not enough retries to create the upload."""

        tus_server["retries_create"] = 11

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)

        location = await aiotus.upload(
            tus_server["create_endpoint"], memory_file, config=config
        )

        assert location is None
        assert tus_server["data"] is None  # Upload could not be created.

    async def test_upload_upload_retries_exceeded(self, tus_server, memory_file):
        """Not enough retries to do the upload."""

        tus_server["retries_upload"] = 11

        config = aiotus.RetryConfiguration(max_retry_period_seconds=0.001)

        location = await aiotus.upload(
            tus_server["create_endpoint"], memory_file, config=config
        )

        assert location is None
        assert tus_server["data"] is not None  # Upload could be created.

    async def test_upload_relative_create(self, tus_server, memory_file):
        """Test what happens if the server returns a relative URL on creation."""

        # Make the create-handler return only the last part of the URL.
        path = tus_server["upload_endpoint"].path
        tus_server["upload_endpoint"] = yarl.URL(path.split("/")[-1])

        location = await aiotus.upload(tus_server["create_endpoint"], memory_file)

        assert location is not None
        assert tus_server["data"] is not None
        assert tus_server["data"] == memory_file.getbuffer()

    async def test_upload_file(self, tus_server):
        """Test upload of an actual file, not an 'io.BytesIO'."""

        md1 = {"key1": "value1".encode(), "key2": "value2".encode()}

        with open(__file__, "rb") as file:
            location = await aiotus.upload(tus_server["create_endpoint"], file, md1)

        assert location is not None
        assert tus_server["data"] is not None
        with open(__file__, "rb") as file:
            data = file.read()
            assert tus_server["data"] == data

        md2 = await aiotus.metadata(str(location))
        assert md1 == md2

    async def test_tusd(self, tusd, memory_file):
        """Test communication with the the tusd server."""

        data = memory_file.getvalue()

        location = await aiotus.upload(tusd.url, memory_file)

        async with aiohttp.ClientSession() as session:
            async with session.get(location) as response:
                body = await response.read()

                assert body == data

        md = await aiotus.metadata(str(location))
        assert not md  # We did not upload any metadata.

        # Upload a file with metadata to tusd, and check if we can read it back.
        md1 = {"key1": "value1".encode(), "key2": "value2".encode()}
        location = await aiotus.upload(tusd.url, memory_file, md1)

        md2 = await aiotus.metadata(location)
        assert md1 == md2



"""Test the implementation of the creation extension."""

import aiohttp
import pytest  # type: ignore

import aiotus


class TestCreate:
    async def test_create_wrong_metadata(self, memory_file):
        """Check the different checks performed on metadata keys."""

        with pytest.raises(ValueError) as excinfo:
            metadata = {"k1": "v1".encode(), "k²": "v2".encode(), "k3": "v3".encode()}
            await aiotus.creation.create(None, None, memory_file, metadata)

        assert "ASCII characters" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            metadata = {"k1": "v1".encode(), "k 2": "v2".encode(), "k3": "v3".encode()}
            await aiotus.creation.create(None, None, memory_file, metadata)

        assert "spaces" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            metadata = {"k1": "v1".encode(), "k2,": "v2".encode(), "k3": "v3".encode()}
            await aiotus.creation.create(None, None, memory_file, metadata)

        assert "commas" in str(excinfo.value)

    async def test_create_wrong_status(self, aiohttp_server, memory_file):
        """Check if status code is checked correctly."""

        async def handler_wrong_status(request):
            raise aiohttp.web.HTTPBadRequest()

        app = aiohttp.web.Application()
        app.router.add_route("POST", "/wrong_status", handler_wrong_status)
        server = await aiohttp_server(app)

        with pytest.raises(aiotus.ProtocolError) as excinfo:
            endpoint = server.make_url("/wrong_status")

            async with aiohttp.ClientSession() as session:
                await aiotus.creation.create(session, endpoint, memory_file, {})

        assert "Wrong status code" in str(excinfo.value)

    async def test_create_no_location(self, aiohttp_server, memory_file):
        """Check if the check for the "Location" header is working."""

        async def handler_no_location(request):
            raise aiohttp.web.HTTPCreated()

        app = aiohttp.web.Application()
        app.router.add_route("POST", "/no_location", handler_no_location)
        server = await aiohttp_server(app)

        with pytest.raises(RuntimeError) as excinfo:
            endpoint = server.make_url("/no_location")

            async with aiohttp.ClientSession() as session:
                await aiotus.creation.create(session, endpoint, memory_file, {})

        assert 'no "Location" header' in str(excinfo.value)

    async def test_create_functional(self, tus_server, memory_file):
        """Test the normal functionality of the upload creation."""

        metadata = {"k1": "1".encode(), "k2": "2²".encode(), "k-3": "three".encode()}

        endpoint = tus_server["server"].make_url("/files")

        async with aiohttp.ClientSession() as session:
            location = await aiotus.creation.create(
                session, endpoint, memory_file, metadata
            )

        assert tus_server["upload_endpoint"] == location



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiotus",
    version="0.1.3",
    author="Jens Steinhauser",
    author_email="jens.steinhauser@gmail.com",
    description="Asynchronous tus (tus.io) client library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JenSte/aiotus",
    project_urls={"Documentation": "https://aiotus.readthedocs.io"},
    license="Apache 2",
    packages=setuptools.find_packages(),
    package_data={"aiotus": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "aiotus-upload=aiotus.entrypoint:aiotus_upload",
            "aiotus-metadata=aiotus.entrypoint:aiotus_metadata",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Typing :: Typed",
    ],
    python_requires=">=3.7",
    install_requires=["aiohttp", "tenacity"],
    setup_requires=["setuptools_scm"],
    use_scm_version={"local_scheme": "dirty-tag"},
)



"""Test the entrypoint module."""

import io
import os.path
import sys
import unittest.mock

import aiotus.entrypoint


class TestAiotusClients:
    def test_aiotus_clients(self, tusd):
        with unittest.mock.patch(
            "sys.argv", ["aiotus-upload", "--debug", str(tusd.url) + "x", __file__]
        ):
            assert 1 == aiotus.entrypoint.aiotus_upload()

        with unittest.mock.patch(
            "sys.argv", ["aiotus-upload", str(tusd.url), __file__ + "x"]
        ):
            assert 1 == aiotus.entrypoint.aiotus_upload()

        with unittest.mock.patch(
            "sys.argv", ["aiotus-upload", str(tusd.url), __file__]
        ):
            with unittest.mock.patch("sys.stdout", new_callable=io.StringIO):
                assert 0 == aiotus.entrypoint.aiotus_upload()
                url = sys.stdout.getvalue().strip()

        expected_output = [
            f"filename: {os.path.basename(__file__)}",
            "mime_type: text/x-python",
        ]

        with unittest.mock.patch("sys.argv", ["aiotus-metadata", url]):
            with unittest.mock.patch("sys.stdout", new_callable=io.StringIO):
                assert 0 == aiotus.entrypoint.aiotus_metadata()

                lines = sys.stdout.getvalue().splitlines(False)
                lines.sort()

                assert lines == expected_output

        with unittest.mock.patch("sys.argv", ["aiotus-metadata", "--debug", url]):
            with unittest.mock.patch("sys.stdout", new_callable=io.StringIO):
                assert 0 == aiotus.entrypoint.aiotus_metadata()

                lines = sys.stdout.getvalue().splitlines(False)
                lines.sort()

                assert lines == expected_output
